/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.revision.helper;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipInputStream;

import org.LexGrid.LexBIG.Impl.logging.LgLoggerIF;
import org.LexGrid.util.sql.DBUtility;
import org.LexGrid.util.sql.GenericSQLModifier;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.LexGrid.util.sql.lgTables.SQLTableUtilities;
import org.LexGrid.versions.Revision;
import org.lexgrid.extension.revision.exception.LBRevisionException;
import org.lexgrid.extension.revision.persistence.LgRevisionConstants;

public class LgRevisionDAO {

	private Connection connection_ = null;
	private String tablePrefix_ = null;
	private SQLTableUtilities sqlTableUtil_ = null;
	private HashMap<String, String> attributeMap_ = null;
	private StringBuffer whereClause_ = null;
	private LgLoggerIF message_ = null;

	public LgRevisionDAO(Connection conn, String tablePrefix, LgLoggerIF message) {

		connection_ = conn;
		tablePrefix_ = tablePrefix;
		message_ = message;

		try {
			sqlTableUtil_ = new SQLTableUtilities(connection_, tablePrefix_);
		} catch (Exception e) {
			message_.fatal("LgRevisionDAO initialization exception", e);
		}
	}

/**
	 * comparator for sorting in descending order of the time.
	 */
@SuppressWarnings("unchecked")
	Comparator<Timestamp> dateComparator = new Comparator() {

		public int compare(Object o1, Object o2) {
			if (o1 instanceof Timestamp && o2 instanceof Timestamp) {
				Timestamp t1 = (Timestamp) o1;
				Timestamp t2 = (Timestamp) o2;

				return t2.compareTo(t1);
			}
			return 1;
		}
	};

	/**
	 * Method returns the id for the latest revision on which the given entity
	 * was modified.
	 * 
	 * @param codingSchemeName
	 * @param namespace
	 * @param entityCode
	 * @return
	 * @throws LBRevisionException
	 */
	public String getLatestEntityRevisionId(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		String revisionId = null;

		Integer entryStateId = getEntryStateIdforEntity(codingSchemeName,
				entityCode);

		if (entryStateId != null)
			revisionId = getRevisionId(entryStateId);

		return revisionId;
	}

	/**
	 * Method returns the oldest(first) revision id for the given entity.
	 * 
	 * @param codingSchemeName
	 * @param entityCode
	 * @return
	 * @throws LBRevisionException
	 */
	public String getOldestEntityRevisionId(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		List<String> allRevIdForEntity = getAllRevIdsForEntity(
				codingSchemeName, entityCode);

		String oldestRev = null;

		int revCount = allRevIdForEntity.size();

		if (revCount != 0) {
			oldestRev = (String) allRevIdForEntity.get(revCount - 1);
		}

		return oldestRev;
	}

	public String getLatestPropertyRevisionId(String codingSchemeName,
			String entityCode, String propertyId) {

		int entryStateId = getEntryStateIdforProperty(codingSchemeName,
				entityCode, propertyId);

		String revisionId = getRevisionId(entryStateId);

		return revisionId;
	}

	public Integer getEntryStateIdforEntity(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		attributeMap_ = new HashMap<String, String>();
		attributeMap_.put("1", SQLTableConstants.TBLCOL_ENTRYSTATEID);

		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_CODINGSCHEMENAME + " = ");
		whereClause_.append("\'");
		whereClause_.append(codingSchemeName);
		whereClause_.append("\'");

		whereClause_.append(" AND ");

		whereClause_.append(SQLTableConstants.TBLCOL_ENTITYCODE + " = ");
		whereClause_.append("\'");
		whereClause_.append(entityCode);
		whereClause_.append("\'");

		ResultSet result = null;
		Integer entryStateId = null;

		try {

			result = sqlTableUtil_.extractDataFromDB(
					SQLTableConstants.TBL_ENTITY, attributeMap_, whereClause_
							.toString(), connection_.getMetaData()
							.getDatabaseProductName());

			if (result.next()) {
				entryStateId = result
						.getInt(SQLTableConstants.TBLCOL_ENTRYSTATEID);
			} else {
				throw new LBRevisionException("The entity " + entityCode
						+ " is either invalid or has"
						+ " been removed from the CodingScheme.");
			}

		} catch (SQLException e) {
			message_.error("Exception @ getEntryStateIdforEntity method", e);
		}

		return entryStateId;
	}

	public boolean isValidRevisionId(String revisionId) {

		attributeMap_ = new HashMap<String, String>();
		
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REVISIONID);

		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_REVISIONID + " = ");
		whereClause_.append("\'");
		whereClause_.append(revisionId);
		whereClause_.append("\'");
		
		try {

			ResultSet result = sqlTableUtil_.extractDataFromDB(
					LgRevisionConstants.TBL_REVISION, attributeMap_, whereClause_.toString(),
					connection_.getMetaData().getDatabaseProductName());
			
			if (result.next()) {
				return true;
			}

		} catch (SQLException e) {
			message_.error("Exception @ isValidRevisionId method", e);
		}

		return false;
	}

	public Timestamp getRevisionDate(String revisionId) {

		Timestamp revisionDate = null;

		attributeMap_ = new HashMap<String, String>();
		
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REVISIONDATE);
		
		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_REVISIONID + " = ");
		whereClause_.append("\'");
		whereClause_.append(revisionId);
		whereClause_.append("\'");

		try {

			ResultSet result = sqlTableUtil_.extractDataFromDB(
					LgRevisionConstants.TBL_REVISION, attributeMap_, whereClause_.toString(),
					connection_.getMetaData().getDatabaseProductName());
			
			if (result.next()) {
				revisionDate = result
						.getTimestamp(LgRevisionConstants.TBLCOL_REVISIONDATE);
				
				return revisionDate;
			}

		} catch (SQLException e) {
			message_.error("Exception @ getRevisionDate method", e);
		}

		return null;

	}
	
	/**
	 * Returns the revisionDate and its corresponding revisionId for all the
	 * revisions in the datebase. The dates are sorted in descending order.
	 * 
	 * @return Map
	 */
	public Map<Timestamp, String> getAllRevisionDate() {

		Map<Timestamp, String> revisionList = new TreeMap<Timestamp, String>(
				dateComparator);

		String revisionID = null;
		Timestamp revisionDate = null;

		attributeMap_ = new HashMap<String, String>();
		
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REVISIONID);
		attributeMap_.put("2", LgRevisionConstants.TBLCOL_REVISIONDATE);

		try {

			ResultSet result = sqlTableUtil_.extractDataFromDB(
					LgRevisionConstants.TBL_REVISION, attributeMap_, null,
					connection_.getMetaData().getDatabaseProductName());
			
			while (result.next()) {
				revisionDate = result
						.getTimestamp(LgRevisionConstants.TBLCOL_REVISIONDATE);
				revisionID = result
						.getString(LgRevisionConstants.TBLCOL_REVISIONID);
				revisionList.put(revisionDate, revisionID);
			}

		} catch (SQLException e) {
			message_.error("Exception @ getAllRevisionDate method", e);
		}

		return revisionList;
	}

	/**
	 * Returns the revAppliedDate and its corresponding revisionId for all the
	 * revisions in the datebase. The dates are sorted in descending order.
	 * 
	 * @return Map
	 */
	public Map<Timestamp, String> getAllRevisionAppliedDate() {

		Map<Timestamp, String> revisionList = new TreeMap<Timestamp, String>(
				dateComparator);

		String revisionID = null;
		Timestamp appliedDate = null;

		attributeMap_ = new HashMap<String, String>();
		
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REVISIONID);
		attributeMap_.put("2", LgRevisionConstants.TBLCOL_REV_APPLIED_DATE);

		try {

			ResultSet result = sqlTableUtil_.extractDataFromDB(
					LgRevisionConstants.TBL_REVISION, attributeMap_, null,
					connection_.getMetaData().getDatabaseProductName());
			
			while (result.next()) {
				appliedDate = result
						.getTimestamp(LgRevisionConstants.TBLCOL_REV_APPLIED_DATE);
				revisionID = result
						.getString(LgRevisionConstants.TBLCOL_REVISIONID);
				revisionList.put(appliedDate, revisionID);
			}

		} catch (SQLException e) {
			message_.error("Exception @ getAllRevisionAppliedDate method", e);
		}

		return revisionList;
	}
	
	/**
	 * Method returns all the revisions available for the codingscheme.
	 * Revisions are listed in descending order of the revisionAppliedDate
	 * (latest first).
	 * 
	 * @return revisionIds
	 */
	public List<String> getAllRevisionIds() {

		List<String> revisionIds = new ArrayList<String>();

		Map<Timestamp, String> revisionDate = getAllRevisionAppliedDate();

		Iterator<Timestamp> itr = revisionDate.keySet().iterator();
		while (itr.hasNext()) {
			String revId = (String) revisionDate.get(itr.next());
			revisionIds.add(revId);
		}

		return revisionIds;
	}
	
	/**
	 * Returns the revAppliedDate for the given revision.
	 * 
	 * @return Timestamp : Revision applied date.
	 */
	public Timestamp getRevisionAppliedDate(String revisionId) {

		Timestamp appliedDate = null;
		
		if( revisionId == null ) {
			return null;
		}

		attributeMap_ = new HashMap<String, String>();
		
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REV_APPLIED_DATE);
		
		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_REVISIONID + " = ");
		whereClause_.append("\'");
		whereClause_.append(revisionId);
		whereClause_.append("\'");

		try {

			ResultSet result = sqlTableUtil_.extractDataFromDB(
					LgRevisionConstants.TBL_REVISION, attributeMap_, whereClause_.toString(),
					connection_.getMetaData().getDatabaseProductName());
			
			if (result.next()) {
				appliedDate = result
						.getTimestamp(LgRevisionConstants.TBLCOL_REV_APPLIED_DATE);
				
				return appliedDate;
			}

		} catch (SQLException e) {
			message_.error("Exception @ getRevisionAppliedDate method", e);
		}

		return null;
	}

	/**
	 * Method returns effective date for the entity.
	 * 
	 * @param codingSchemeName
	 * @param entityCode
	 * @return
	 * @throws LBRevisionException
	 */
	public Timestamp getEntityEffectiveDate(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		String revisionId = getLatestEntityRevisionId(codingSchemeName,
				entityCode);

		Timestamp effectiveDate = getEntityEffectiveDate(codingSchemeName,
				entityCode, revisionId);

		return effectiveDate;
	}

	public Timestamp getEntityEffectiveDate(String codingSchemeName,
			String entityCode, String revisionId) throws LBRevisionException {

		Integer entryStateId = getEntryStateIdforEntity(codingSchemeName,
				entityCode);
		Timestamp effDate = null;
		Integer prevEntryStateId = new Integer(-1);

		while (prevEntryStateId != null) {

			attributeMap_ = new HashMap<String, String>();
			attributeMap_.put("1", SQLTableConstants.TBLCOL_REVISIONID);
			attributeMap_.put("2", SQLTableConstants.TBLCOL_EFFECTIVEDATE);
			attributeMap_.put("3", SQLTableConstants.TBLCOL_PREVENTRYSTATEID);

			whereClause_ = new StringBuffer();

			whereClause_.append(SQLTableConstants.TBLCOL_ENTRYSTATEID + " = ");
			whereClause_.append(entryStateId);

			ResultSet results = null;

			try {

				results = sqlTableUtil_.extractDataFromDB(
						SQLTableConstants.TBL_ENTRY_STATE, attributeMap_,
						whereClause_.toString(), connection_.getMetaData()
								.getDatabaseProductName());

				if( results.next() ) {
					String revId = results
							.getString(SQLTableConstants.TBLCOL_REVISIONID);
	
					if (revisionId == null
							|| (revId != null && revId.equals(revisionId))) {
						effDate = results
								.getTimestamp(SQLTableConstants.TBLCOL_EFFECTIVEDATE);
	
						return effDate;
					}
	
					prevEntryStateId = results
							.getInt(SQLTableConstants.TBLCOL_PREVENTRYSTATEID);
					
					entryStateId = prevEntryStateId;
				} else {
					prevEntryStateId = null;
				}
				
				

			} catch (SQLException e) {
				message_.error("Exception @ getEntityEffectiveDate method", e);
			}
		}
		return effDate;
	}
	
	/**
	 * Method returns expiry date for the entity.
	 * 
	 * @param codingSchemeName
	 * @param entityCode
	 * @return
	 * @throws LBRevisionException
	 */
	public Timestamp getEntityExpiryDate(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		String revisionId = getLatestEntityRevisionId(codingSchemeName,
				entityCode);

		Timestamp expiryDate = getEntityExpiryDate(codingSchemeName,
				entityCode, revisionId);

		return expiryDate;
	}

	public Timestamp getEntityExpiryDate(String codingSchemeName,
			String entityCode, String revisionId) throws LBRevisionException {

		Integer entryStateId = getEntryStateIdforEntity(codingSchemeName,
				entityCode);
		Timestamp expDate = null;
		Integer prevEntryStateId = new Integer(-1);

		while (prevEntryStateId != null) {

			attributeMap_ = new HashMap<String, String>();
			attributeMap_.put("1", SQLTableConstants.TBLCOL_REVISIONID);
			attributeMap_.put("2", SQLTableConstants.TBLCOL_EXPIRATIONDATE);
			attributeMap_.put("3", SQLTableConstants.TBLCOL_PREVENTRYSTATEID);

			whereClause_ = new StringBuffer();

			whereClause_.append(SQLTableConstants.TBLCOL_ENTRYSTATEID + " = ");
			whereClause_.append(entryStateId);

			ResultSet results = null;

			try {

				results = sqlTableUtil_.extractDataFromDB(
						SQLTableConstants.TBL_ENTRY_STATE, attributeMap_,
						whereClause_.toString(), connection_.getMetaData()
								.getDatabaseProductName());

				if (results.next()) {
					String revId = results
							.getString(SQLTableConstants.TBLCOL_REVISIONID);

					if (revisionId == null
							|| (revId != null && revId.equals(revisionId))) {
						expDate = results
								.getTimestamp(SQLTableConstants.TBLCOL_EXPIRATIONDATE);

						return expDate;
					}

					prevEntryStateId = results
							.getInt(SQLTableConstants.TBLCOL_PREVENTRYSTATEID);
					entryStateId = prevEntryStateId;
					
				} else {
					prevEntryStateId = null;
				}

			} catch (SQLException e) {
				message_.error("Exception @ getEntityExpiryDate method", e);
			}
		}
		return expDate;
	}

	@SuppressWarnings("unchecked")
	public Revision getRevisionData(String revisionId) {

		ResultSet result = null;
		Boolean isZipped = null;
		Blob revisionData = null;
		Revision revision = null;

		String query = "SELECT " + LgRevisionConstants.TBLCOL_ISZIPPED + ", "
				+ LgRevisionConstants.TBLCOL_REVISIONDATA + " FROM "
				+ tablePrefix_ + LgRevisionConstants.TBL_REVISION
				+ " WHERE " + LgRevisionConstants.TBLCOL_REVISIONID + " = "
				+ "\'" + revisionId + "\'";

		try {

			PreparedStatement prepStmt = connection_.prepareStatement(query);
			
			result = prepStmt.executeQuery();

			if (result.next()) {

				isZipped = DBUtility.getBooleanFromResultSet(result,
						LgRevisionConstants.TBLCOL_ISZIPPED);
				
				InputStream byteStream = null;
				
				if( connection_.getMetaData().getDatabaseProductName().equals("PostgreSQL")) {
					byte[] revData = result
						.getBytes(LgRevisionConstants.TBLCOL_REVISIONDATA);
					
					byteStream = new ByteArrayInputStream(revData);
				} else {
					revisionData = result
							.getBlob(LgRevisionConstants.TBLCOL_REVISIONDATA);
					
					byteStream = revisionData.getBinaryStream();
				}
				
				org.exolab.castor.xml.Unmarshaller um = new org.exolab.castor.xml.Unmarshaller(
						Revision.class);

				if (isZipped != null && isZipped.booleanValue()) {

					ZipInputStream in = new ZipInputStream(byteStream);
					in.getNextEntry();

					revision = (Revision) um
							.unmarshal(new InputStreamReader(in));

				} else {
					revision = (Revision) um.unmarshal(new InputStreamReader(
							byteStream));
				}
			}

		} catch (Exception e) {
			message_.error("Exception @ getRevisionData method", e);
		}

		return revision;
	}
	
	@SuppressWarnings("unchecked")
	public Revision getPrevRevisionData(String revisionId) {

		ResultSet result = null;
		Boolean isZipped = null;
		Blob revisionData = null;
		Revision revision = null;

		String query = "SELECT " + LgRevisionConstants.TBLCOL_ISZIPPED + ", "
				+ LgRevisionConstants.TBLCOL_PREVREVISIONDATA + " FROM "
				+ tablePrefix_ + LgRevisionConstants.TBL_REVISION
				+ " WHERE " + LgRevisionConstants.TBLCOL_REVISIONID + " = "
				+ "\'" + revisionId + "\'";
		
		try {

			PreparedStatement prepStmt = connection_.prepareStatement(query);
			
			result = prepStmt.executeQuery();

			if (result.next()) {

				isZipped = DBUtility.getBooleanFromResultSet(result,
						LgRevisionConstants.TBLCOL_ISZIPPED);
				
				InputStream byteStream = null;
				
				if (connection_.getMetaData().getDatabaseProductName().equals(
						"PostgreSQL")) {
					byte[] revData = result
							.getBytes(LgRevisionConstants.TBLCOL_PREVREVISIONDATA);

					byteStream = new ByteArrayInputStream(revData);
				} else {
					revisionData = result
							.getBlob(LgRevisionConstants.TBLCOL_PREVREVISIONDATA);

					byteStream = revisionData.getBinaryStream();
				}

				org.exolab.castor.xml.Unmarshaller um = new org.exolab.castor.xml.Unmarshaller(
						Revision.class);

				if (isZipped != null && isZipped.booleanValue()) {

					ZipInputStream in = new ZipInputStream(byteStream);
					in.getNextEntry();

					revision = (Revision) um
							.unmarshal(new InputStreamReader(in));
				} else {
					revision = (Revision) um.unmarshal(new InputStreamReader(
							byteStream));
				}
			}

		} catch (Exception e) {
			message_.error("Exception @ getPrevRevisionData method", e);
		}

		return revision;
	}

	/**
	 * - Method calculates the closest earlier revision to the given revision
	 * ('revisionId') on which entityCode was modified. - EntityCode may not be
	 * modified on revisionId. - If the codingScheme or entityCode was never
	 * revised, method returns revisionId itself. - If revisionId is an
	 * arbitrary id (invalid revision id) method throws LBRevisionException. -If
	 * revisionId is older than the oldest revision id of entityCode, method
	 * returns null.
	 * 
	 * @param codingSchemeName
	 * @param entityCode
	 * @param revisionId
	 * @return
	 * @throws LBRevisionException
	 */
	@SuppressWarnings("unchecked")
	public String getPrevRevIDforEntity(String codingSchemeName,
			String entityCode, String revisionId) throws LBRevisionException {

		Timestamp revisionAppDate = getRevisionAppliedDate(revisionId);

		if (revisionAppDate == null) {
			throw new LBRevisionException("Invalid Revision Id : " + revisionId);
		}

		List<String> entityRevIdList = getAllRevIdsForEntity(codingSchemeName,
				entityCode);

		for (String revId : entityRevIdList) {

			Timestamp revAppDate = getRevisionAppliedDate(revId);

			if (revAppDate.before(revisionAppDate)) {
				return revId;
			}
		}

		return null;
	}

	/**
	 * Method returns the list of all the revision id's on which the given
	 * entity was modified.
	 * 
	 * @param codingSchemeName
	 * @param entityCode
	 * @return
	 * @throws LBRevisionException
	 */
	public List<String> getAllRevIdsForEntity(String codingSchemeName,
			String entityCode) throws LBRevisionException {

		List<String> entityRevIdList = new ArrayList<String>();

		Integer entryStateId = getEntryStateIdforEntity(codingSchemeName,
				entityCode);

		Integer prevEntryStateId = new Integer(-1);

		while (prevEntryStateId != null) {

			attributeMap_ = new HashMap<String, String>();
			attributeMap_.put("1", SQLTableConstants.TBLCOL_REVISIONID);
			attributeMap_.put("2", SQLTableConstants.TBLCOL_PREVENTRYSTATEID);

			whereClause_ = new StringBuffer();

			whereClause_.append(SQLTableConstants.TBLCOL_ENTRYSTATEID + " = ");
			whereClause_.append(entryStateId);

			ResultSet results = null;

			try {

				results = sqlTableUtil_.extractDataFromDB(
						SQLTableConstants.TBL_ENTRY_STATE, attributeMap_,
						whereClause_.toString(), connection_.getMetaData()
								.getDatabaseProductName());

				if (results.next()) {
					String revId = results
							.getString(SQLTableConstants.TBLCOL_REVISIONID);

					if (revId != null)
						entityRevIdList.add(revId);

					prevEntryStateId = results
							.getInt(SQLTableConstants.TBLCOL_PREVENTRYSTATEID);

					entryStateId = prevEntryStateId;
				} else {
					prevEntryStateId = null;
				}

			} catch (SQLException e) {
				message_.error("Exception @ getAllRevIdsForEntity method", e);
			}
		}

		return entityRevIdList;
	}

	/**
	 * Method returns true if revision1 and revision2 are equal or revisionDate
	 * of revision1 is after revisionDate of revision2. Returns null if either
	 * revision1 or revision2 are invalid revisions.
	 * 
	 * @param revisionId
	 * @param latestRevisionId
	 * @return
	 */
	public Boolean isLatestRevision(String revision1, String revision2) {

		if (!isValidRevisionId(revision1) || !isValidRevisionId(revision2)) {
			return null;
		}

		if (revision1.equals(revision2))
			return true;

		if (getRevisionDate(revision1).after(getRevisionDate(revision2))) {
			return true;
		}

		return false;
	}

	private int getEntryStateIdforProperty(String codingSchemeName,
			String entityCode, String propertyId) {

		attributeMap_ = new HashMap<String, String>();
		attributeMap_.put("1", SQLTableConstants.TBLCOL_ENTRYSTATEID);

		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_CODINGSCHEMENAME + " = ");
		whereClause_.append("\'");
		whereClause_.append(codingSchemeName);
		whereClause_.append("\'");

		whereClause_.append(" AND ");

		whereClause_.append(SQLTableConstants.TBLCOL_ENTITYCODE + " = ");
		whereClause_.append("\'");
		whereClause_.append(entityCode);
		whereClause_.append("\'");

		whereClause_.append(" AND ");

		whereClause_.append(SQLTableConstants.TBLCOL_PROPERTYID + " = ");
		whereClause_.append("\'");
		whereClause_.append(propertyId);
		whereClause_.append("\'");

		ResultSet result = null;
		int entryStateId = -1;

		try {

			result = sqlTableUtil_.extractDataFromDB(
					SQLTableConstants.TBL_ENTITY_PROPERTY, attributeMap_,
					whereClause_.toString(), connection_.getMetaData()
							.getDatabaseProductName());

			if (result.next()) {
				entryStateId = result
						.getInt(SQLTableConstants.TBLCOL_ENTRYSTATEID);
			}

		} catch (SQLException e) {
			message_.error("Exception @ getEntryStateIdforProperty method", e);
		}

		return entryStateId;
	}

	private String getRevisionId(int entryStateId) {

		attributeMap_ = new HashMap<String, String>();
		attributeMap_.put("1", LgRevisionConstants.TBLCOL_REVISIONID);

		whereClause_ = new StringBuffer();

		whereClause_.append(SQLTableConstants.TBLCOL_ENTRYSTATEID + " = ");
		whereClause_.append(entryStateId);

		ResultSet result = null;
		String revisionId = null;

		try {

			result = sqlTableUtil_.extractDataFromDB(
					SQLTableConstants.TBL_ENTRY_STATE, attributeMap_,
					whereClause_.toString(), connection_.getMetaData()
							.getDatabaseProductName());

			if (result.next()) {
				revisionId = result
						.getString(LgRevisionConstants.TBLCOL_REVISIONID);
			}

		} catch (SQLException e) {
			message_.error("Exception @ getRevisionId method", e);
		}

		return revisionId;
	}
	
    /**
	 * Runs SQL Statement "SELECT" with supplied attributes and where clause
	 * 
	 * @param tableName
	 * @param attributeNames
	 * @param whereClause
	 * @return
	 * @throws SQLException
	 */
	public ResultSet extractDataFromDB(String tableName, Map attributeNames,
			String whereClause, String dbType) throws SQLException {

		StringBuffer stmt = new StringBuffer();
		PreparedStatement prepStmt = null;
		ResultSet resultSet = null;

		stmt.append("SELECT ");

		for (int i = 0; i < attributeNames.size(); i++) {
			stmt.append(attributeNames.get("" + (i + 1)) + ",");
		}

		stmt = stmt.deleteCharAt(stmt.length() - 1);

		stmt.append(" FROM ");
		stmt.append(tablePrefix_ + tableName);

		if (whereClause != null && !whereClause.equals("")) {
			stmt.append(" WHERE ");
			stmt.append(whereClause);
		}

		message_.debug("************ SELECT QUERY ************");
		message_.debug(stmt.toString());
		message_.debug("**************************************");

		try {
			String statement = new GenericSQLModifier(dbType, false)
					.modifySQL(stmt.toString());

			prepStmt = connection_.prepareStatement(statement);
			resultSet = prepStmt.executeQuery();
		} catch (Exception e) {
			message_.error("Exception @ extractDataFromDB: " + e.getMessage());
		} finally {
			// prepStmt.close();
		}

		return resultSet;
	}
}