/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.revision.helper;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.commonTypes.EntityDescription;
import org.LexGrid.commonTypes.Property;
import org.LexGrid.commonTypes.PropertyQualifier;
import org.LexGrid.commonTypes.Source;
import org.LexGrid.commonTypes.Text;
import org.LexGrid.concepts.Comment;
import org.LexGrid.concepts.Definition;
import org.LexGrid.concepts.Entities;
import org.LexGrid.concepts.Entity;
import org.LexGrid.concepts.Presentation;
import org.LexGrid.concepts.PropertyLink;
import org.LexGrid.versions.ChangedEntry;
import org.LexGrid.versions.EntryState;
import org.LexGrid.versions.Revision;
import org.LexGrid.versions.types.ChangeType;
import org.lexgrid.extension.revision.exception.LBRevisionException;

public class LgRevisionUtil {

	public static Map<Long, Entity> getEntityFromRevision(Revision revision,
			String codingSchemeName, String version, String entityCode) {

		Entities entities = null;
		Entity entity = null;
		Map<Long, Entity> entityMap = new TreeMap<Long, Entity>();

		if (revision != null) {

			ChangedEntry[] changedEntries = revision.getChangedEntry();

			if (changedEntries != null) {
				for (int i = 0; i < changedEntries.length; i++) {
					CodingScheme codingScheme = changedEntries[i]
							.getChangedCodingSchemeEntry();

					if (codingScheme != null
							&& codingScheme.getCodingSchemeName().equals(
									codingSchemeName)) {

						// version check.
						if (version != null) {

							String repVersion = codingScheme
									.getRepresentsVersion();
							if (!version.equals(repVersion)) {
								continue;
							}
						}

						entities = codingScheme.getEntities();

						if (entities != null) {
							for (int j = 0; j < entities.getEntityCount(); j++) {

								entity = entities.getEntity(j);
								String tempEntityCode = entity.getEntityCode();
								EntryState entryState = entity.getEntryState();
								Long order = null;

								if (entryState != null) {
									order = entryState.getRelativeOrder();
								}

								if (entityCode.equals(tempEntityCode)) {
									if (order != null) {
										entityMap.put(order, entity);
									} else {
										entityMap.put(new Long(0), entity);
										break;
									}
								}
							}

							return entityMap;
						}
					}
				}
			}
		}

		return null;
	}

	public static CodingScheme getCodingSchemeFromRevision(Revision revision,
			String codingSchemeName, String version, String entityCode) {

		if (revision != null) {

			ChangedEntry[] changedEntries = revision.getChangedEntry();

			if (changedEntries != null) {
				for (int i = 0; i < changedEntries.length; i++) {
					CodingScheme codingScheme = changedEntries[i]
							.getChangedCodingSchemeEntry();

					if (codingScheme != null
							&& codingScheme.getCodingSchemeName().equals(
									codingSchemeName)) {

						// version check.
						if (version != null) {

							String repVersion = codingScheme
									.getRepresentsVersion();
							if (!version.equals(repVersion)) {
								continue;
							}
						}

						return codingScheme;
					}
				}
			}
		}

		return null;
	}

	/**
	 * Method applies changes to the prevEntity. Changes are applied in the same
	 * order as entityChanges.
	 * 
	 * @param prevEntity
	 * @param entityChanges
	 * @return
	 * @throws LBRevisionException
	 */
	public static Entity applyEntityChanges(Entity prevEntity,
			Map<Long, Entity> entityChanges) throws LBRevisionException {

		Iterator<Long> itr = entityChanges.keySet().iterator();

		while (itr.hasNext() && prevEntity != null) {

			Long order = (Long) itr.next();

			Entity changes = (Entity) entityChanges.get(order);

			prevEntity = applyEntityChanges(prevEntity, changes);
		}

		return prevEntity;
	}

	/**
	 * Method applies 'changes' to 'prevEntity'.
	 * 
	 * @param prevEntity
	 * @param changes
	 * @return
	 * @throws LBRevisionException
	 */
	public static Entity applyEntityChanges(Entity prevEntity, Entity changes)
			throws LBRevisionException {

		EntryState entryState = changes.getEntryState();

		if (entryState != null) {

			ChangeType cType = entryState.getChangeType();

			if (cType != null) {
				if (cType.equals(ChangeType.NEW)) {
					return changes;

				} else if (cType.equals(ChangeType.REMOVE)) {

					throw new LBRevisionException(
							"The entity "
									+ changes.getEntityCode()
									+ " was removed from the codingscheme in this revision.");

				} else if (cType.equals(ChangeType.MODIFY)) {

					if (prevEntity == null) {
						throw new LBRevisionException("Invalid entity data.");
					}

					Boolean isActive = changes.getIsActive();
					Boolean isAnonymous = changes.getIsAnonymous();
					Boolean isDefined = changes.getIsDefined();
					EntryState eState = changes.getEntryState();
					EntityDescription entityDesc = changes
							.getEntityDescription();
					String[] eTypes = changes.getEntityType();
					Date effDate = changes.getEffectiveDate();
					Date expDate = changes.getExpirationDate();
					String status = changes.getStatus();
					Source owner = changes.getOwner();

					if (isActive != null)
						prevEntity.setIsActive(isActive);

					if (isAnonymous != null)
						prevEntity.setIsAnonymous(isAnonymous);

					if (isDefined != null)
						prevEntity.setIsDefined(isDefined);

					if (effDate != null)
						prevEntity.setEffectiveDate(effDate);

					if (expDate != null)
						prevEntity.setExpirationDate(expDate);

					if (eState != null)
						prevEntity.setEntryState(eState);

					if (entityDesc != null)
						prevEntity.setEntityDescription(entityDesc);

					if (eTypes != null)
						prevEntity.setEntityType(eTypes);

					if (status != null)
						prevEntity.setStatus(status);

					if (owner != null)
						prevEntity.setOwner(owner);

					prevEntity = applyPropertyChanges(prevEntity, changes);

				} else if (cType.equals(ChangeType.DEPENDENT)) {

					prevEntity = applyPropertyChanges(prevEntity, changes);

				} else if (cType.equals(ChangeType.VERSIONABLE)) {

					Boolean isActive = changes.getIsActive();
					Source owner = changes.getOwner();
					String status = changes.getStatus();
					Date effDate = changes.getEffectiveDate();
					Date expDate = changes.getExpirationDate();

					if (isActive != null)
						prevEntity.setIsActive(isActive);
					if (owner != null)
						prevEntity.setOwner(owner);
					if (status != null)
						prevEntity.setStatus(status);
					if (effDate != null)
						prevEntity.setEffectiveDate(effDate);
					if (expDate != null)
						prevEntity.setExpirationDate(expDate);
				}
			}
		}
		return prevEntity;
	}

	public static Entity applyPropertyChanges(Entity prevEntity,
			Entity entityChanges) {

		/* ------------- */

		Property[] allPrevProps = prevEntity.getAllProperties();

		List<Property> allPrevPropsList = arrayToList(allPrevProps);

		Map<String, Property> prevPropMap = getPropIdToPropertyMap(allPrevProps);

		/* ------------- */

		Property[] allPropChanges = entityChanges.getAllProperties();

		Map<Long, Property> allPropChangesMap = sortPropertyByRelativeOrder(allPropChanges);

		/* ------------- */

		PropertyLink[] propLinkChanges = entityChanges.getPropertyLink();

		List<PropertyLink> allPropLinkChangesList = arrayToList(propLinkChanges);

		PropertyLink[] allPrevPropLink = prevEntity.getPropertyLink();

		List<PropertyLink> allPrevPropLinkList = arrayToList(allPrevPropLink);

		/* ------------- */

		Property propChange = null;
		Property prevProp = null;

		Iterator<Long> itr = allPropChangesMap.keySet().iterator();

		while (itr.hasNext()) {

			Long order = itr.next();
			propChange = allPropChangesMap.get(order);
			String propId = propChange.getPropertyId();
			EntryState eState = propChange.getEntryState();

			if (eState != null) {
				ChangeType cType = eState.getChangeType();

				if (cType != null) {

					prevProp = null;

					if (prevPropMap.containsKey(propId)) {
						prevProp = prevPropMap.get(propId);
					}

					if (cType.equals(ChangeType.NEW)) {
						if (prevProp != null) {
							System.out.println("Duplicate property "
									+ propChange.getPropertyId()
									+ ". Corrupted data. Neglecting "
									+ propChange.getPropertyId());
							continue;
						}
					} else if (prevProp == null) {
						System.out.println("Property "
								+ propChange.getPropertyId()
								+ ". Corrupted data. Neglecting "
								+ propChange.getPropertyId());
						continue;
					}

					if (cType.equals(ChangeType.NEW)) {

						Boolean isActive = propChange.getIsActive();
						if (isActive == null) {
							propChange.setIsActive(true);
						}

						if (propChange instanceof Presentation) {
							Boolean isPreffered = ((Presentation) propChange)
									.getIsPreferred();

							if (isPreffered == null) {
								((Presentation) propChange)
										.setIsPreferred(false);
							}
						} else if (propChange instanceof Definition) {
							Boolean isPreffered = ((Definition) propChange)
									.getIsPreferred();

							if (isPreffered == null) {
								((Definition) propChange).setIsPreferred(false);
							}
						}

						allPrevPropsList.add(propChange);

					} else if (cType.equals(ChangeType.REMOVE)) {

						allPrevPropsList.remove(prevProp);

						if (allPrevPropLinkList != null) {
							for (int j = 0; j < allPrevPropLinkList.size(); j++) {
								PropertyLink propLink = allPrevPropLinkList
										.get(j);
								if (propLink.getSourceProperty().equals(propId)
										|| propLink.getTargetProperty().equals(
												propId)) {
									allPrevPropLinkList.remove(j);
								}
							}
						}

					} else if (cType.equals(ChangeType.MODIFY)) {

						String propName = propChange.getPropertyName();
						String propType = propChange.getPropertyType();
						String propStatus = propChange.getStatus();
						Source[] propSource = propChange.getSource();
						String[] propUsageContext = propChange
								.getUsageContext();
						Text propValue = propChange.getValue();
						String propLang = propChange.getLanguage();
						PropertyQualifier[] propQualifier = propChange
								.getPropertyQualifier();
						Boolean isActive = propChange.getIsActive();
						EntryState entryState = propChange.getEntryState();
						Source owner = propChange.getOwner();
						Date propEffDate = propChange.getEffectiveDate();
						Date propExpDate = propChange.getExpirationDate();

						if (propName != null)
							prevProp.setPropertyName(propName);
						if (propType != null)
							prevProp.setPropertyType(propType);
						if (propValue != null)
							prevProp.setValue(propValue);
						if (propLang != null)
							prevProp.setLanguage(propLang);
						if (propStatus != null)
							prevProp.setStatus(propStatus);
						if (isActive != null)
							prevProp.setIsActive(isActive);
						if (entryState != null)
							prevProp.setEntryState(entryState);
						if (owner != null)
							prevProp.setOwner(owner);
						if (propEffDate != null)
							prevProp.setEffectiveDate(propEffDate);
						if (propExpDate != null)
							prevProp.setExpirationDate(propExpDate);

						if (propSource != null) {

							Source[] prevSource = prevProp.getSource();
							List<Source> prevSourceList = arrayToList(prevSource);

							if (prevSource == null || prevSource.length == 0) {
								prevProp.setSource(propSource);
							} else {

								List<Source> propChangeList = arrayToList(propSource);

								for (int i = 0; i < propChangeList.size(); i++) {
									Source newSource = propChangeList.get(i);

									if (!sourceAlreadyExist(prevSourceList,
											newSource)) {
										prevSourceList.add(newSource);
									}
								}

								prevProp.setSource(prevSourceList
										.toArray(new Source[prevSourceList
												.size()]));
							}
						}

						if (propUsageContext != null) {

							String[] prevUContext = prevProp.getUsageContext();
							List<String> prevUContextList = arrayToList(prevUContext);

							if (prevUContext == null
									|| prevUContext.length == 0) {
								prevProp.setUsageContext(propUsageContext);
							} else {

								List<String> propUContextList = arrayToList(propUsageContext);

								for (int i = 0; i < propUContextList.size(); i++) {
									String newUContext = propUContextList
											.get(i);

									if (!propUContextList.contains(newUContext)) {
										prevUContextList.add(newUContext);
									}
								}

								prevProp.setUsageContext(propUContextList
										.toArray(new String[propUContextList
												.size()]));
							}
						}

						if (propQualifier != null) {

							PropertyQualifier[] prevPropQual = prevProp
									.getPropertyQualifier();
							List<PropertyQualifier> prevPropQualList = arrayToList(prevPropQual);

							if (prevPropQual == null
									|| prevPropQual.length == 0) {
								prevProp.setPropertyQualifier(propQualifier);
							} else {

								List<PropertyQualifier> propQualList = arrayToList(propQualifier);

								for (int i = 0; i < propQualList.size(); i++) {
									PropertyQualifier newPropQual = propQualList
											.get(i);

									if (!propertyQualifierAlreadyExist(
											prevPropQualList, newPropQual)) {
										prevPropQualList.add(newPropQual);
									}
								}

								prevProp
										.setPropertyQualifier((PropertyQualifier[]) propQualList
												.toArray(new PropertyQualifier[propQualList
														.size()]));
							}
						}

						if (propChange instanceof Presentation) {

							Presentation prop = (Presentation) propChange;

							Boolean isPref = prop.getIsPreferred();
							String degreeOfFidelity = prop
									.getDegreeOfFidelity();
							Boolean matchIfNoContext = prop
									.getMatchIfNoContext();
							String repForm = prop.getRepresentationalForm();

							if (isPref != null)
								((Presentation) prevProp)
										.setIsPreferred(isPref);
							if (degreeOfFidelity != null)
								((Presentation) prevProp)
										.setDegreeOfFidelity(degreeOfFidelity);
							if (matchIfNoContext != null)
								((Presentation) prevProp)
										.setMatchIfNoContext(matchIfNoContext);
							if (repForm != null)
								((Presentation) prevProp)
										.setRepresentationalForm(repForm);

						} else if (propChange instanceof Definition) {

							Definition prop = (Definition) propChange;

							Boolean isPref = prop.getIsPreferred();

							if (isPref != null)
								((Definition) prevProp).setIsPreferred(isPref);
						}

					} else if (cType.equals(ChangeType.VERSIONABLE)) {

						Boolean isActive = propChange.getIsActive();
						Date effDate = propChange.getEffectiveDate();
						Date expDate = propChange.getExpirationDate();
						String status = propChange.getStatus();
						Source owner = propChange.getOwner();
						EntryState entryState = propChange.getEntryState();

						if (isActive != null)
							prevProp.setIsActive(isActive);
						if (effDate != null)
							prevProp.setEffectiveDate(effDate);
						if (expDate != null)
							prevProp.setExpirationDate(expDate);
						if (status != null)
							prevProp.setStatus(status);
						if (owner != null)
							prevProp.setOwner(owner);
						if (entryState != null)
							prevProp.setEntryState(entryState);
					}
				}
			}
		}

		/* -------------------- */

		if (allPrevPropsList != null && allPrevPropsList.size() != 0) {

			prevEntity.removeAllComment();
			prevEntity.removeAllDefinition();
			prevEntity.removeAllPresentation();
			prevEntity.removeAllProperty();
			prevEntity.removeAllPropertyLink();

			for (int i = 0; i < allPrevPropsList.size(); i++) {

				if (allPrevPropsList.get(i) instanceof Presentation) {

					prevEntity.addPresentation((Presentation) allPrevPropsList
							.get(i));

				} else if (allPrevPropsList.get(i) instanceof Definition) {

					prevEntity.addDefinition((Definition) allPrevPropsList
							.get(i));

				} else if (allPrevPropsList.get(i) instanceof Comment) {

					prevEntity.addComment((Comment) allPrevPropsList.get(i));

				} else if (allPrevPropsList.get(i) instanceof Property) {

					prevEntity.addProperty((Property) allPrevPropsList.get(i));
				}
			}

			allPrevPropLinkList = addPropertyLink(allPrevPropLinkList,
					allPropLinkChangesList);

			if (allPrevPropLinkList != null && allPrevPropLinkList.size() != 0) {
				prevEntity.setPropertyLink(allPrevPropLinkList
						.toArray(new PropertyLink[allPrevPropLinkList.size()]));
			}
		}

		return prevEntity;
	}

	public static List<PropertyLink> addPropertyLink(
			List<PropertyLink> allPrevPropLinkList,
			List<PropertyLink> allPropLinkChangesList) {

		if (allPropLinkChangesList != null) {
			for (int i = 0; i < allPropLinkChangesList.size(); i++) {
				PropertyLink propLink = allPropLinkChangesList.get(i);
				boolean propLinkExist = false;

				if (propLink != null) {
					String source = propLink.getSourceProperty();
					String link = propLink.getPropertyLink();
					String target = propLink.getTargetProperty();

					if (allPrevPropLinkList != null) {
						for (int j = 0; j < allPrevPropLinkList.size(); j++) {
							PropertyLink prevPropLink = allPrevPropLinkList
									.get(j);
							String prevSource = prevPropLink
									.getSourceProperty();
							String prevLink = prevPropLink.getPropertyLink();
							String prevTarget = prevPropLink
									.getTargetProperty();

							if (source.equals(prevSource)
									&& link.equals(prevLink)
									&& target.equals(prevTarget)) {
								propLinkExist = true;
								break;
							}
						}
					}

					if (!propLinkExist) {
						allPrevPropLinkList.add(propLink);
					}
				}
			}
		}

		return allPrevPropLinkList;
	}

	public static Boolean isEntityRemovedInRevision(Revision revision,
			String codingSchemeName, String version, String entityCode) {

		Map<Long, Entity> prevEntityMap = LgRevisionUtil.getEntityFromRevision(
				revision, codingSchemeName, version, entityCode);

		if (prevEntityMap != null && prevEntityMap.size() != 0) {

			Iterator<Long> itr = prevEntityMap.keySet().iterator();

			while (itr.hasNext()) {
				Long order = (Long) itr.next();
				Entity prevEnt = (Entity) prevEntityMap.get(order);

				EntryState entryState = prevEnt.getEntryState();

				if (entryState != null) {
					ChangeType changeType = entryState.getChangeType();

					if (changeType != null
							&& ChangeType.REMOVE.equals(changeType)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	public static <T> List<T> arrayToList(T[] objArray) {

		List<T> objList = null;

		if (objArray != null) {
			objList = new ArrayList<T>();
			for (int i = 0; i < objArray.length; i++) {
				objList.add(objArray[i]);
			}
		}

		return objList;
	}
	
	private static boolean sourceAlreadyExist(List<Source> prevSourceList,
			Source newSource) {

		if (prevSourceList != null) {
			for (int i = 0; i < prevSourceList.size(); i++) {
				Source prevSource = prevSourceList.get(i);

				String subRef = prevSource.getSubRef();
				String value = prevSource.getContent();

				if (subRef.equals(newSource.getSubRef())
						&& value.equals(newSource.getContent())) {
					return true;
				}
			}
		}

		return false;
	}

	private static boolean propertyQualifierAlreadyExist(
			List<PropertyQualifier> prevPropQualList,
			PropertyQualifier newPropQual) {

		if (prevPropQualList != null) {
			for (int i = 0; i < prevPropQualList.size(); i++) {
				PropertyQualifier prevPropQual = prevPropQualList.get(i);

				String qualName = prevPropQual.getPropertyQualifierName();
				String qualValue = prevPropQual.getValue().getContent();

				if (qualName.equals(newPropQual.getPropertyQualifierName())
						&& qualValue
								.equals(newPropQual.getValue().getContent())) {
					return true;
				}
			}
		}

		return false;
	}

	private static Map<String, Property> getPropIdToPropertyMap(
			Property[] allProps) {

		Map<String, Property> propertyIdToPropMap = new HashMap<String, Property>();

		for (int j = 0; j < allProps.length; j++) {

			Property prop = allProps[j];
			if (prop != null) {
				String prevPropId = prop.getPropertyId();

				propertyIdToPropMap.put(prevPropId, prop);
			}
		}

		return propertyIdToPropMap;
	}

	private static Map<Long, Property> sortPropertyByRelativeOrder(
			Property[] allPropChanges) {

		Map<Long, Property> allPropChangesMap = new TreeMap<Long, Property>();

		for (int i = 0; i < allPropChanges.length; i++) {

			Long order = allPropChanges[i].getEntryState().getRelativeOrder();

			allPropChangesMap.put(order, allPropChanges[i]);
		}

		return allPropChangesMap;
	}
}