/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.revision.impl;

import java.sql.Connection;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.LexGrid.LexBIG.DataModel.Collections.ResolvedConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Impl.LexBIGServiceImpl;
import org.LexGrid.LexBIG.Impl.dataAccess.ResourceManager;
import org.LexGrid.LexBIG.Impl.dataAccess.SQLInterface;
import org.LexGrid.LexBIG.Impl.internalExceptions.MissingResourceException;
import org.LexGrid.LexBIG.Impl.logging.Logger;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.concepts.Entity;
import org.LexGrid.managedobj.jdbc.JDBCConnectionDescriptor;
import org.LexGrid.util.sql.DBUtility;
import org.LexGrid.versions.EntryState;
import org.LexGrid.versions.Revision;
import org.LexGrid.versions.types.ChangeType;
import org.lexgrid.extension.revision.LgRevisedCodedNodeSet;
import org.lexgrid.extension.revision.exception.LBRevisionException;
import org.lexgrid.extension.revision.helper.LgRevisionDAO;
import org.lexgrid.extension.revision.helper.LgRevisionUtil;

/**
 * @author m057089
 * 
 */
public class LgRevisedCodedNodeSetImpl implements LgRevisedCodedNodeSet {

	/** */
	private String codingSchemeName_ = null;
	/** */
	private CodingSchemeVersionOrTag tagOrVersion_ = null;
	/** */
	private Connection connection_ = null;
	/** */
	private String tablePrefix_ = null;
	/** */
	private String version_ = null;
	/** */
	private Logger message_ = null;

	/**
	 * 
	 * @param codingScheme
	 * @param tagOrVersion
	 * @param message_
	 * @param activeOnly
	 * @throws LBRevisionException 
	 */
	public LgRevisedCodedNodeSetImpl(String codingScheme,
			CodingSchemeVersionOrTag tagOrVersion, Logger message) throws LBRevisionException {

		tagOrVersion_ = tagOrVersion;
		message_ = message;

		if (tagOrVersion_ == null) {
			try {
				version_ = ResourceManager.instance()
						.getInternalVersionStringFor(codingScheme, null);
			} catch (LBParameterException e) {
				throw new LBRevisionException(
						"Version Initialization exception",
						e);
			}
		} else {
			version_ = tagOrVersion_.getVersion();
		}

		try {
			codingSchemeName_ = ResourceManager.instance()
					.getInternalCodingSchemeNameForUserCodingSchemeName(
							codingScheme, version_);
		} catch (LBParameterException e) {
			throw new LBRevisionException(
					"LgRevisedCodedNodeSetImpl - CodingSchemeName Initialization exception",
					e);
		}

		initConnection();
	}

	/**
	 * This method resolves the concept reference for the given entity
	 * ('entityCode') at a given point of time ('date'). Method iterates through
	 * the different revisions of the entity and returns the latest revision of
	 * the entity available on 'date' within LexBIG system. If the given entity
	 * is 'REMOVE'ed from the lexBIG system, method throws LBRevisionException.
	 * If 'date' is null, method resolves the concept for current system
	 * date.Method can'r resolve concepts for future dates.
	 * 
	 * @param entityCode
	 * @param date
	 */
	public ResolvedConceptReference resolveConceptByDate(String entityCode,
			Date date) throws LBRevisionException {

		ResolvedConceptReference resConceptRef = null;
		String revisionId = null;

		if (date == null) {
			date = new Date(System.currentTimeMillis());
		} else if (date.after(new Date(System.currentTimeMillis()))) {

			throw new LBRevisionException(
					"Concept cant't be resolved for future dates");
		}

		LgRevisionDAO revisionDAO = new LgRevisionDAO(getConnection(),
				getTablePrefix(), message_);

		// If the entity is deleted, getAllRevIdsForEntity will throw
		// LBRevisionException and so stops further processing and returns
		// from the method.
		List<String> entityRevIdList = revisionDAO.getAllRevIdsForEntity(
				codingSchemeName_, entityCode);

		// If the codingScheme has no revisions or the entity is never revised
		// just resolve the concept using core lexbig api.
		if (entityRevIdList.size() == 0) {

			resConceptRef = resolveConcept(entityCode);
			
			return resConceptRef;
		}

		for (int i = 0; i < entityRevIdList.size(); i++) {

			revisionId = entityRevIdList.get(i);
			Timestamp revAppliedTime = revisionDAO
					.getRevisionAppliedDate(revisionId);

			if (date.after(revAppliedTime)) {

				resConceptRef = resolveConceptByRevision(entityCode, revisionId);
				
				return resConceptRef;
			}
		}

		Entity initialLoad = null;

		String oldestRevId = entityRevIdList.get(entityRevIdList.size() - 1);

		Revision revData = revisionDAO.getPrevRevisionData(oldestRevId);

		CodingScheme codingScheme = LgRevisionUtil.getCodingSchemeFromRevision(
				revData, codingSchemeName_, version_, entityCode);

		Map<Long, Entity> revDataMap = LgRevisionUtil.getEntityFromRevision(
				revData, codingSchemeName_, version_, entityCode);

		Iterator<Long> itr = revDataMap.keySet().iterator();

		if (itr.hasNext()) {
			initialLoad = revDataMap.get(itr.next());
		}

		if (initialLoad != null) {

			resConceptRef = getResolvedConceptReference(codingScheme,
					initialLoad);
		}
		
		return resConceptRef;
	}

	/**
	 * This method resolves the concept reference for the given entity
	 * ('entityCode') at a given revision ('revisionId'). Method iterates
	 * through the different revisions of the entity and returns the state of the 
	 * entity (values of attributes and properties)on given revision. If
	 * the given entity is 'REMOVE'ed from the lexBIG system, method throws
	 * LBRevisionException.
	 * 
	 * @param entityCode
	 * @param revisionId
	 */
	public ResolvedConceptReference resolveConceptByRevision(String entityCode,
			String revisionId) throws LBRevisionException {

		ResolvedConceptReference resConceptRef = null;

		LgRevisionDAO revisionDAO = new LgRevisionDAO(getConnection(),
				getTablePrefix(), message_);

		if (revisionId != null && !revisionDAO.isValidRevisionId(revisionId)) {
			throw new LBRevisionException("Invalid revision : " + revisionId);
		}

		// If the entity is deleted, getAllRevIdsForEntity will throw
		// LBRevisionException and so stops further processing and return
		// from the method.
		List<String> allRevIdsForEntity = revisionDAO.getAllRevIdsForEntity(
				codingSchemeName_, entityCode);

		// check if the entity was ever revised?
		if (allRevIdsForEntity.size() == 0 || revisionId == null) {
			return resolveConcept(entityCode);
		}
		
		String latestRevisionId = (String) allRevIdsForEntity.get(0);

		// check if revisionId is same or newer than the latest revision of
		// entityCode
		
		if (revisionDAO.isLatestRevision(revisionId, latestRevisionId)) {
			return resolveConcept(entityCode);
		}
		
		if (!allRevIdsForEntity.contains(revisionId)) {

			int revCount = allRevIdsForEntity.size();

			String oldestRevisionId = (String) allRevIdsForEntity
					.get(revCount - 1);

			// check if revisionId is same or older than the oldest revision of
			// entityCode

			if (revisionDAO.isLatestRevision(oldestRevisionId, revisionId)) {

				Revision oldestRevision = revisionDAO
						.getRevisionData(oldestRevisionId);

				Map<Long, Entity> oldEntityMap = LgRevisionUtil
						.getEntityFromRevision(oldestRevision,
								codingSchemeName_, version_, entityCode);

				if (oldEntityMap != null && oldEntityMap.size() != 0) {

					Iterator<Entity> itr = oldEntityMap.values().iterator();

					if (itr.hasNext()) {

						Entity oldEntity = (Entity) itr.next();

						ChangeType changeType = oldEntity.getEntryState()
								.getChangeType();

						if (changeType != null
								&& ChangeType.NEW.equals(changeType)) {

							throw new LBRevisionException("The entity "
									+ entityCode + " didn't exist in revision "
									+ revisionId);
						} else {
							Revision oldestOrigData = revisionDAO
									.getPrevRevisionData(oldestRevisionId);

							CodingScheme codingScheme = LgRevisionUtil
									.getCodingSchemeFromRevision(
											oldestOrigData, codingSchemeName_,
											version_, entityCode);

							Map<Long, Entity> oldOrigEntityMap = LgRevisionUtil
									.getEntityFromRevision(oldestOrigData,
											codingSchemeName_, version_,
											entityCode);
							
							Iterator<Long> entMapItr = oldOrigEntityMap
									.keySet().iterator();

							if (entMapItr.hasNext()) {

								Long order = entMapItr.next();

								return getResolvedConceptReference(
										codingScheme, oldOrigEntityMap
												.get(order));
							} else {
								throw new LBRevisionException(
										"Incorrect data encountered for entity "
												+ entityCode);
							}
						}
					}
				}
				
			} else {
				String prevRevisionId = revisionDAO.getPrevRevIDforEntity(
						codingSchemeName_, entityCode, revisionId);

				revisionId = prevRevisionId;
			}
		}

		resConceptRef = applyRevisionChanges(entityCode, revisionId);

		return resConceptRef;
	}

	private ResolvedConceptReference applyRevisionChanges(String entityCode,
			String revisionId) throws LBRevisionException {

		Map<Long, Entity> entityChanges = null;
		Map<Long, Entity> prevEntity = null;
		ResolvedConceptReference resConceptRef = null;

		LgRevisionDAO revisionDAO = new LgRevisionDAO(getConnection(),
				getTablePrefix(), message_);

		Revision revisionData = revisionDAO.getRevisionData(revisionId);
		Revision prevRevisionData = revisionDAO.getPrevRevisionData(revisionId);

		CodingScheme codingScheme = LgRevisionUtil.getCodingSchemeFromRevision(
				revisionData, codingSchemeName_, version_, entityCode);

		entityChanges = LgRevisionUtil.getEntityFromRevision(revisionData,
				codingSchemeName_, version_, entityCode);

		prevEntity = LgRevisionUtil.getEntityFromRevision(prevRevisionData,
				codingSchemeName_, version_, entityCode);

		if (entityChanges != null) {

			Iterator<Long> itr = entityChanges.keySet().iterator();

			if (itr.hasNext()) {
				Long key = (Long) itr.next();
				Entity ent = entityChanges.get(key);

				EntryState entrySt = ent.getEntryState();

				if (entrySt != null) {
					if (ChangeType.NEW.equals(entrySt.getChangeType())) {

						prevEntity = new TreeMap<Long, Entity>();

						prevEntity.put(0L, ent);
						entityChanges.remove(key);
					}
				}
			}

			if (prevEntity != null) {

				Iterator<Long> itrPrev = prevEntity.keySet().iterator();

				if (itrPrev.hasNext()) {
					Long key = itrPrev.next();

					Entity pEntity = prevEntity.get(key);

					Entity changedEntity = LgRevisionUtil.applyEntityChanges(
							pEntity, entityChanges);

					if (changedEntity != null) {
						resConceptRef = getResolvedConceptReference(
								codingScheme, changedEntity);
					}
				}
			}
		}
		return resConceptRef;
	}

	private ResolvedConceptReference getResolvedConceptReference(
			CodingScheme codingScheme, Entity entity) {

		ResolvedConceptReference conceptRef = null;

		if (codingScheme != null && entity != null) {
			conceptRef = new ResolvedConceptReference();

			conceptRef.setCode(entity.getEntityCode());
			conceptRef.setCodeNamespace(entity.getEntityCodeNamespace());
			conceptRef.setCodingSchemeName(codingScheme.getCodingSchemeName());
			conceptRef.setCodingSchemeURI(codingScheme.getCodingSchemeURI());
			conceptRef.setCodingSchemeVersion(codingScheme
					.getRepresentsVersion());
			conceptRef.setConceptCode(entity.getEntityCode());
			conceptRef.setEntity(entity);
			conceptRef.setEntityDescription(entity.getEntityDescription());
			conceptRef.setEntityType(entity.getEntityType());
			conceptRef.setSourceOf(null);
			conceptRef.setTargetOf(null);
		}

		return conceptRef;
	}

	private ResolvedConceptReference resolveConcept(String entityCode)
			throws LBRevisionException {

		CodedNodeSet codedNodeSet = null;
		ResolvedConceptReferenceList rcrl = null;
		LexBIGService lbSvc = LexBIGServiceImpl.defaultInstance();

		try {
			codedNodeSet = lbSvc.getCodingSchemeConcepts(codingSchemeName_,
					tagOrVersion_);

			codedNodeSet.restrictToCodes(Constructors
					.createConceptReferenceList(entityCode, codingSchemeName_));

			rcrl = codedNodeSet.resolveToList(null, null, null, 1);
		} catch (LBException e) {

			throw new LBRevisionException("Exception while resolving concept.",
					e);
		}

		if (rcrl != null && rcrl.getResolvedConceptReferenceCount() != 0)
			return rcrl.getResolvedConceptReference(0);
		else
			return null;
	}

	private Connection getConnection() {
		return connection_;
	}

	private String getTablePrefix() {
		return tablePrefix_;
	}

	private void initConnection() throws LBRevisionException {
		SQLInterface sqlInterface = null;

		try {
			sqlInterface = ResourceManager.instance().getSQLInterface(
					codingSchemeName_, version_);
		} catch (MissingResourceException e) {
			message_.fatal("Couldn't get handle for the codingScheme : "
					+ codingSchemeName_ + ", Version : " + version_);
			throw new LBRevisionException(
					"Couldn't get handle for the codingScheme : "
							+ codingSchemeName_ + ", Version : " + version_, e);
		}

		JDBCConnectionDescriptor connDesc = sqlInterface
				.getConnectionDescriptor();
		String server = connDesc.getDbUrl();
		String driver = connDesc.getDbDriver();
		String user = connDesc.getDbUid();
		String password = connDesc.getDbPwd();

		tablePrefix_ = sqlInterface.getTablePrefix() == null ? ""
				: sqlInterface.getTablePrefix();

		try {
			connection_ = DBUtility.connectToDatabase(server, driver, user,
					password);
		} catch (Exception e) {
			message_.fatal("Database connection failure");
			throw new LBRevisionException("Database connection failure", e);
		}
	}
}