/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.revision.persistence;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;

import org.LexGrid.util.sql.GenericSQLModifier;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.apache.log4j.Logger;
import org.lexgrid.extension.revision.exception.LBRevisionException;

public class LgRevisionSQLService {

	private static Logger log = Logger.getLogger("LgRevisionSQLService.SQL");
	private Connection sqlConnection_;
	private GenericObjectPool connectionPool_;
	private String tablePrefix_;
	private GenericSQLModifier gsm_;
	private Hashtable<String, String> tableCreateSql_ = new Hashtable<String, String>();
	private Hashtable<String, String> insertStatements_ = new Hashtable<String, String>();
	private ArrayList<String> indexSql_ = new ArrayList<String>();
	private ArrayList<String> foreignKeySql_ = new ArrayList<String>();

	public static final String versionString = "1.7";
	public static final String tableStructureDescription = "This is version 1.7 of the LexGrid SQL format - this is compatible with the 2009/01 LexGrid Schema";

	public LgRevisionSQLService(Connection sqlConnection, String tablePrefix)
			throws LBRevisionException {
		sqlConnection_ = sqlConnection;
		connectionPool_ = null;
		tablePrefix_ = tablePrefix;
		try {
			gsm_ = new GenericSQLModifier(sqlConnection_);
		} catch (Exception e) {
			throw new LBRevisionException(
					"LgRevisionSQLService Initialization exception.", e);
		}
	}

	/**
	 * Use this constructor if you would rather have it check out connections
	 * from the pool as needed.
	 * 
	 * @param connectionPool
	 * @param tablePrefix
	 * @throws LBRevisionException
	 */
	public LgRevisionSQLService(GenericObjectPool connectionPool,
			String tablePrefix) throws LBRevisionException {
		sqlConnection_ = null;
		connectionPool_ = connectionPool;
		tablePrefix_ = tablePrefix;

		Connection conn = getConnection();
		try {
			gsm_ = new GenericSQLModifier(conn);
		} catch (Exception e) {

			throw new LBRevisionException(
					"LgRevisionSQLService Initialization exception.", e);
		} finally {
			returnConnection(conn);
		}
	}

	public String getExistingRevisionTableVersion() throws LBRevisionException {
		PreparedStatement checkVersion = null;

		if (!doRevisionTablesExist()) {
			log.debug("Revision tables don't exist - returning.");
			return null;
		}

		String result = "<1.5";

		Connection conn = getConnection();

		try {
			checkVersion = conn
					.prepareStatement("Select "
							+ SQLTableConstants.TBLCOL_VERSION
							+ " from "
							+ getTableName(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA));
			ResultSet results = checkVersion.executeQuery();
			if (results.next()) {
				result = results.getString(SQLTableConstants.TBLCOL_VERSION);
			}
			results.close();
			return result;

		} catch (SQLException e) {
			return "<1.5";
		} finally {
			try {
				if (checkVersion != null) {
					checkVersion.close();
				}
			} catch (SQLException e) {
				// do nothing
			}
			returnConnection(conn);
		}
	}

	/**
	 * Returns a specific statement based on a key
	 * 
	 * @param key
	 *            the search key
	 * @return the sql statement
	 * @throws LBRevisionException
	 */
	public String getRevisioningInsertStatementSQL(String key)
			throws LBRevisionException {
		if (insertStatements_.size() == 0)
			initInsertStatements();

		key = getTableName(key);
		String t = (String) insertStatements_.get(key);
		if (t == null) {
			throw new LBRevisionException(
					"Insert statement not available for the table '" + key
							+ "'");
		} else {
			return t;
		}
	}

	public boolean doRevisionTablesExist() throws LBRevisionException {
		boolean exist = true;
		exist = exist
				&& doesTableExist(LgRevisionConstants.TBL_SYSTEM_RELELEASE);
		exist = exist && doesTableExist(LgRevisionConstants.TBL_REVISION);

		return exist;
	}

	public boolean doesTableExist(String tableName) throws LBRevisionException {
		PreparedStatement query = null;

		Connection conn = getConnection();

		try {
			query = conn.prepareStatement("Select count(*) from "
					+ getTableName(tableName));
			ResultSet results = query.executeQuery();
			if (results.next()) {
				// if the table doesn't exist, it should throw an exception.
				results.close();
				return true;
			} else {
				return false; // should not be used.
			}
		} catch (SQLException e1) {
			return false;
		} finally {
			try {
				if (query != null) {
					query.close();
				}
			} catch (SQLException e) {
				// do nothing
			}
			returnConnection(conn);
		}
	}

	/**
	 * Initializes and creates revision tables
	 * 
	 * @throws LBRevisionException
	 */
	public void createRevisionTables() throws LBRevisionException {

		if (doRevisionTablesExist()) {
			log.debug("Revision tables already exist, not creating new ones.");
			return;
		}

		if (tableCreateSql_.size() == 0) {
			log.debug("initiating revision table creation sql");
			initTableCreateSQL();
		}

		createRevisionMetaDataTable();

		Connection conn = getConnection();
		try {

			Enumeration<String> tempEnum = tableCreateSql_.keys();
			while (tempEnum.hasMoreElements()) {
				String key = (String) tempEnum.nextElement();
				log.debug("Creating " + key + " table");
				createTable(conn, gsm_.modifySQL((String) tableCreateSql_
						.get(key)), key);
			}
		} finally {
			returnConnection(conn);
		}

		createRevisionTableIndexes();
		createRevisionTableConstraints();
	}

	public void deleteRevisionTables() throws LBRevisionException {

		if (doesTableExist(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA)) {
			dropTable(getConnection(),
					LgRevisionConstants.TBL_REVISION_TABLE_META_DATA);
		}

		if (doesTableExist(LgRevisionConstants.TBL_SYSTEM_RELELEASE)) {
			dropTable(getConnection(), LgRevisionConstants.TBL_SYSTEM_RELELEASE);
		}

		if (doesTableExist(LgRevisionConstants.TBL_REVISION)) {
			dropTable(getConnection(), LgRevisionConstants.TBL_REVISION);
		}
	}

	private void dropTable(Connection conn, String tableName) {
		PreparedStatement delete = null;
		try {
			delete = conn.prepareStatement(gsm_.modifySQL("DROP TABLE "
					+ getTableName(tableName) + " {CASCADE} "));
			delete.executeUpdate();
			delete.close();
		} catch (SQLException e) {
			log.info(
					"Failed while dropping the table - it probably didn't exist."
							+ getTableName(tableName), e);
		} finally {
			if (delete != null) {
				try {
					delete.close();
				} catch (SQLException e) {
					// noop
				}
			}
		}
	}

	private void initTableCreateSQL() throws LBRevisionException {

		tableCreateSql_
				.put(
						getTableName(LgRevisionConstants.TBL_SYSTEM_RELELEASE),
						"CREATE TABLE {IF NOT EXISTS} ^"
								+ getTableName(LgRevisionConstants.TBL_SYSTEM_RELELEASE)
								+ "^ (" + " ^"
								+ LgRevisionConstants.TBLCOL_RELEASEURI
								+ "^ {limitedText}(250) NOT NULL," + " ^"
								+ LgRevisionConstants.TBLCOL_RELEASEID
								+ "^ {limitedText}(50) NOT NULL," + " ^"
								+ LgRevisionConstants.TBLCOL_RELEASEDATE
								+ "^ {dateTime} NOT NULL," + " ^"
								+ LgRevisionConstants.TBLCOL_BASEDONRELEASE
								+ "^ {limitedText}(250) default NULL," + " ^"
								+ LgRevisionConstants.TBLCOL_RELEASEAGENCY
								+ "^ {limitedText}(250) default NULL," + " ^"
								+ LgRevisionConstants.TBLCOL_ENTITYDESCRIPTION
								+ "^ {unlimitedText} default NULL,"
								+ " PRIMARY KEY  (^"
								+ LgRevisionConstants.TBLCOL_RELEASEURI + "^)"
								+ ") {TYPE} {lgTableCharSet}");

		tableCreateSql_.put(getTableName(LgRevisionConstants.TBL_REVISION),
				"CREATE TABLE {IF NOT EXISTS} ^"
						+ getTableName(LgRevisionConstants.TBL_REVISION)
						+ "^ (" + " ^" + LgRevisionConstants.TBLCOL_REVISIONID
						+ "^ {limitedText}(50) NOT NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_REVISIONRELEASEURI
						+ "^ {limitedText}(50) NOT NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_CHANGEAGENT
						+ "^ {limitedText}(50) default NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_REVISIONDATE
						+ "^ {dateTime} NOT NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_REV_APPLIED_DATE
						+ "^ {dateTime} NOT NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_EDITORDER
						+ "^ {bigInt} NOT NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_ENTITYDESCRIPTION
						+ "^ {unlimitedText} default NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_CHANGEINSTRUCTIONS
						+ "^ {unlimitedText} default NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_ISZIPPED
						+ "^ {boolean} default NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_REVISIONDATA
						+ "^ {blob} default NULL," + " ^"
						+ LgRevisionConstants.TBLCOL_PREVREVISIONDATA
						+ "^ {blob} default NULL," + " PRIMARY KEY  (^"
						+ LgRevisionConstants.TBLCOL_REVISIONID + "^)"
						+ ") {TYPE} {lgTableCharSet}");
	}

	private void createRevisionMetaDataTable() throws LBRevisionException {

		Connection conn = getConnection();

		try {
			String createTable = gsm_
					.modifySQL("CREATE TABLE {IF NOT EXISTS} ^"
							+ getTableName(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA)
							+ "^ (" + " ^" + SQLTableConstants.TBLCOL_VERSION
							+ "^ {limitedText}(50) NOT NULL," + " ^"
							+ SQLTableConstants.TBLCOL_DESCRIPTION
							+ "^ {limitedText}(255) default NULL"
							+ ") {TYPE} {lgTableCharSet}");

			boolean created = createTable(
					conn,
					createTable,
					getTableName(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA));

			if (created) {
				log.debug("Inserting version identifier");
				PreparedStatement tempInsert = conn
						.prepareStatement(gsm_
								.modifySQL(getRevisioningInsertStatementSQL(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA)));
				tempInsert.setString(1, versionString);
				tempInsert.setString(2, tableStructureDescription);
				tempInsert.executeUpdate();
				tempInsert.close();
			}
		} catch (SQLException e) {

			throw new LBRevisionException(
					"Exception while creating metadata for revision tables.", e);
		} finally {
			returnConnection(conn);
		}
	}

	private void createRevisionTableIndexes() throws LBRevisionException {
		log.debug("Creating table indexes");

		if (indexSql_.size() == 0) {
			log.debug("initing index creation sql");
			initIndexCreateSql();
		}
		Connection conn = getConnection();
		try {
			for (int i = 0; i < indexSql_.size(); i++) {
				createIndex(conn, gsm_.modifySQL((String) indexSql_.get(i)),
						(String) indexSql_.get(i));
			}
		} finally {
			returnConnection(conn);
		}
	}

	private void createIndex(Connection conn, String createIndexString,
			String indexName) throws LBRevisionException {
		PreparedStatement ps = null;
		try {
			ps = conn.prepareStatement(createIndexString);
			ps.execute();
		} catch (SQLException e) {
			// try to figure out what the databases return if the constraint
			// already exists... (mysql
			// does the errno 121 business
			if (e.toString().indexOf("already") == -1
					&& e.toString().indexOf("existing") == -1
					&& e.toString().indexOf("errno: 121") == -1
					&& e.toString().indexOf("-601") == -1
					&& e.toString().indexOf("Duplicate key name") == -1) {
				log.error("Problem creating the index " + createIndexString, e);

				throw new LBRevisionException(
						"Exception while creating the indexes for revision tables.",
						e);

			} else {
				log.debug("The index " + indexName + " already exits");
			}
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					// do nothing
				}
			}
		}
	}

	/**
	 * Inserts table constraints
	 * 
	 * @throws LBRevisionException
	 */
	private void createRevisionTableConstraints() throws LBRevisionException {
		log.debug("Creating revision table constraints");
		if (!doRevisionTablesExist()) {
			log.debug("Tables don't exist - returning.");
			return;
		}

		if (foreignKeySql_.size() == 0) {
			log.debug("initing the addForeignKey sql code");
			initForeigKeyAlterSQL();
		}

		Connection conn = getConnection();
		try {

			for (int i = 0; i < foreignKeySql_.size(); i++) {
				createForeignKey(conn, gsm_.modifySQL((String) foreignKeySql_
						.get(i)), (String) foreignKeySql_.get(i));
			}
		} finally {
			returnConnection(conn);
		}
	}

	private void createForeignKey(Connection conn, String createFKString,
			String FKName) throws LBRevisionException {

		PreparedStatement ps = null;
		try {
			ps = conn.prepareStatement(createFKString);
			ps.execute();
		} catch (SQLException e) {
			// try to figure out what the databases return if the constraint
			// already exists... (mysql
			// does the errno 121 business
			if (e.toString().indexOf("already") == -1
					&& e.toString().indexOf("existing") == -1
					&& e.toString().indexOf("Duplicate") == -1
					&& e.toString().indexOf("-601") == -1
					&& e.toString().indexOf("errno: 121") == -1) {
				log.error("Problem loading the foreign key " + createFKString,
						e);
				throw new LBRevisionException("Exception.", e);
			} else {
				log
						.debug("The foreign constraint " + FKName
								+ " already exits");
			}
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					// Do nothing
				}
			}
		}
	}

	private void initForeigKeyAlterSQL() {
		log.debug("initRevisionTableForeignKey called");

		foreignKeySql_.add("ALTER TABLE ^"
				+ getTableName(LgRevisionConstants.TBL_REVISION)
				+ "^ ADD CONSTRAINT ^fkver1^ FOREIGN KEY (^"
				+ LgRevisionConstants.TBLCOL_REVISIONRELEASEURI
				+ "^) REFERENCES ^"
				+ getTableName(LgRevisionConstants.TBL_SYSTEM_RELELEASE)
				+ "^ (^" + LgRevisionConstants.TBLCOL_RELEASEURI + "^)");
	}

	private void initIndexCreateSql() {

		indexSql_.add("CREATE INDEX ^iver2^ ON ^"
				+ getTableName(LgRevisionConstants.TBL_REVISION) + "^ (^"
				+ LgRevisionConstants.TBLCOL_REVISIONRELEASEURI + "^ ) ");
	}

	/*
	 * returns true if the table was created, false otherwise (already existed)
	 */
	private boolean createTable(Connection conn, String createTableString,
			String tableName) throws LBRevisionException {
		PreparedStatement ps = null;
		try {
			ps = conn.prepareStatement(createTableString);
			ps.execute();
			return true;
		} catch (SQLException e) {
			// If the error is anything other than table already exists, throw
			// an error.
			if (e.toString().indexOf("already exists") == -1
					&& e.toString().indexOf("identical") == -1
					&& e.toString().indexOf("-601") == -1
					&& e.toString().indexOf(
							"already used by an existing object") == -1
					&& e.toString().indexOf("already an object") == -1) {

				log.error("Problem creating the " + tableName
						+ " table. Syntax used=" + createTableString, e);

				throw new LBRevisionException(
						"Exception while creating the revision tables.", e);
			} else {
				log.debug("The table " + tableName
						+ " already appears to exist.");
				return false;
			}
		} finally {
			if (ps != null) {
				try {
					ps.close();
				} catch (SQLException e) {
					// do nothing
				}
			}
		}
	}

	private void initInsertStatements() {
		insertStatements_
				.put(
						getTableName(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA),
						"INSERT INTO "
								+ getTableName(LgRevisionConstants.TBL_REVISION_TABLE_META_DATA)
								+ " (" + SQLTableConstants.TBLCOL_VERSION
								+ ", " + SQLTableConstants.TBLCOL_DESCRIPTION
								+ ") VALUES (?, ?)");

		insertStatements_
				.put(
						getTableName(LgRevisionConstants.TBL_SYSTEM_RELELEASE),
						"INSERT INTO "
								+ getTableName(LgRevisionConstants.TBL_SYSTEM_RELELEASE)
								+ " (" + LgRevisionConstants.TBLCOL_RELEASEURI
								+ ", " + LgRevisionConstants.TBLCOL_RELEASEID
								+ ", " + LgRevisionConstants.TBLCOL_RELEASEDATE
								+ ", "
								+ LgRevisionConstants.TBLCOL_BASEDONRELEASE
								+ ", "
								+ LgRevisionConstants.TBLCOL_RELEASEAGENCY
								+ ", "
								+ LgRevisionConstants.TBLCOL_ENTITYDESCRIPTION
								+ ") VALUES (?, ?, ?, ?, ?, ?)");

		insertStatements_.put(getTableName(LgRevisionConstants.TBL_REVISION),
				"INSERT INTO " + getTableName(LgRevisionConstants.TBL_REVISION)
						+ " (" + LgRevisionConstants.TBLCOL_REVISIONID + ", "
						+ LgRevisionConstants.TBLCOL_REVISIONRELEASEURI + ", "
						+ LgRevisionConstants.TBLCOL_CHANGEAGENT + ", "
						+ LgRevisionConstants.TBLCOL_REVISIONDATE + ", "
						+ LgRevisionConstants.TBLCOL_REV_APPLIED_DATE + ", "
						+ LgRevisionConstants.TBLCOL_EDITORDER + ", "
						+ LgRevisionConstants.TBLCOL_ENTITYDESCRIPTION + ", "
						+ LgRevisionConstants.TBLCOL_CHANGEINSTRUCTIONS + ", "
						+ LgRevisionConstants.TBLCOL_ISZIPPED + ", "
						+ LgRevisionConstants.TBLCOL_REVISIONDATA + ", "
						+ LgRevisionConstants.TBLCOL_PREVREVISIONDATA
						+ ") VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
	}

	private String getTableName(String key) {

		if (tablePrefix_ != null && tablePrefix_.trim().length() > 0) {
			return tablePrefix_ + key;
		}

		return key;
	}

	private Connection getConnection() {
		if (sqlConnection_ != null) {
			return sqlConnection_;
		} else {
			try {
				return (Connection) connectionPool_.borrowObject();
			} catch (Exception e) {
				return null;
			}
		}
	}

	private void returnConnection(Connection connection) {
		if (connectionPool_ != null) {
			try {
				connectionPool_.returnObject(connection);
			} catch (Exception e) {
			}
		}
	}
}